#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions decorators
"""

from __future__ import print_function, division, absolute_import

import os
import time
import inspect
import traceback
import threading
import contextlib
from functools import wraps

import tpDcc
from tpDcc.libs import python
from tpDcc.libs.python import debug


def abstractmethod(fn):
    """
    The decorated function should be overridden by a software specific module.
    """

    def new_fn(*args, **kwargs):
        msg = 'Abstract implementation has not been overridden.'
        mode = os.getenv('RIGLIB_ABSTRACTMETHOD_MODE')
        if mode == 'raise':
            raise NotImplementedError(debug.debug_object_string(fn, msg))
        elif mode == 'warn':
            tpDcc.logger.warning(debug.debug_object_string(fn, msg))
        return fn(*args, **kwargs)

    new_fn.__name__ = fn.__name__
    new_fn.__doc__ = fn.__doc__
    new_fn.__dict__ = fn.__dict__
    return new_fn


def accepts(*types, **kw):
    """
    Function decorator that checks if decorated function arguments are of the expected types
    :param types: Expected types of the inputs to the decorated function. Must specify a type for each parameter
    :param kw: (optional) Specification of 'debug' level ( 0 | 1 | 2 )
    """

    if not kw:
        debug = 1
    else:
        debug = kw['debug']
    try:
        def decorator(f):
            def newf(*args):
                if debug is 0:
                    return f(*args)
                args = list(args)
                if not (len(args[1:]) == len(types)):
                    raise AssertionError
                argtypes = tuple(map(type, args[1:]))
                if argtypes != types:
                    msg = debug.format_message(f.__name__, types, argtypes, 0)
                    if debug is 1:
                        try:
                            for i in range(1, len(args)):
                                args[i] = types[i - 1](args[i])
                        except ValueError as stdmsg:
                            raise ValueError(msg)
                        except TypeError as stdmsg:
                            raise TypeError(msg)
                    elif debug is 2:
                        raise TypeError(msg)
                return f(*args)
            newf.__name__ = f.__name__
            return newf
        return decorator
    except KeyError as key:
        raise KeyError(key + 'is not a valid keyword argument')
    except TypeError as msg:
        raise TypeError(msg)


def returns(ret_type, **kw):
    """
    Function decorator that checks if decorated function return value is of the expected type
    :param retType: Excepted type of the decorated function return value. Must specify type for each parameter
    :param kwargs: (optional) Specification of 'debug' level (0 | 1 | 2)
    """

    try:
        if not kw:
            debug = 1
        else:
            debug = kw['debug']

        def decorator(f):
            def newf(*args):
                result = f(*args)
                if debug is 0:
                    return result
                res_type = type(result)
                if res_type != ret_type:
                    msg = debug.format_message(f.__name__, (ret_type,), (res_type,), 1)
                    if debug is 1:
                        try:
                            result = ret_type(result)
                        except ValueError:
                            raise ValueError
                    elif debug is 2:
                        raise TypeError(msg)
                return result
            newf.__name__ = f.__name__
            return newf
        return decorator
    except KeyError as key:
        raise KeyError(key + 'is not a valid keyword argument')
    except TypeError as msg:
        raise TypeError(msg)


def timer(fn):
    """
    Function decorator for simple timer
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        if python.logger.getEffectiveLevel() == 20:
            # If we are in info mode we disable timers
            res = fn(*args, **kwargs)
        else:
            t1 = time.time()
            res = fn(*args, **kwargs)
            t2 = time.time()

            trace = ''
            try:
                mod = inspect.getmodule(args[0])
                trace += '{0} >>'.format(mod.__name__.split('.')[-1])
            except Exception:
                python.logger.debug('function module inspect failure')

            try:
                cls = args[0].__clas__
                trace += '{0}.'.format(args[0].__clas__.__name__)
            except Exception:
                python.logger.debug('function class inspect failure')

            trace += fn.__name__
            python.logger.debug('Timer : %s: took %0.3f ms' % (trace, (t2 - t1) * 1000.0))
        return res
    return wrapper


def print_elapsed_time(f):
    """
    Function decorator that gets the elapsed time
    :param f: fn, function
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        t = time.time()
        res = f(*args, **kwargs)
        print(f.__name__, time.time() - t)
        return res
    return wrapper


def timestamp(f):
    """
    Function decorator that gets the elapsed time with a more descriptive output
    :param f: fn, function
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        res = f(*args, **kwargs)
        python.logger.info('<{}> Elapsed time : {}'.format(f.func_name, time.time() - start_time))
        return res
    return wrapper


def try_pass(fn):
    """
    Function decorator that tries a function and if it fails pass
    :param fn: function
    """

    def wrapper(*args, **kwargs):
        return_value = None
        try:
            return_value = fn(*args, **kwargs)
        except Exception:
            python.logger.error(traceback.format_exc())
        return return_value
    return wrapper


def empty_decorator(f):
    """
    Empty decorator
    :param f: fn
    """

    def wrapper(*args, **kwargs):
        r = f(*args, **kwargs)
        return r
    return wrapper


@contextlib.contextmanager
def empty_decorator_context():
    """
    Empty decorator
    :param f: fn
    """

    pass


def repeater(interval, limit=-1):
    """!
    A function interval decorator based on
    http://stackoverflow.com/questions/5179467/equivalent-of-setinterval-in-python

    Inifinite Example Usage:
        @repeater(.05)
        def infinite():
            print "Executing infinity."

        import time
        print "Starting Infinite Repeating Function"
        token = infinite()
        time.sleep(1)
        print "Stopping: infinite"
        token.stop()

    Limited Example Usage:
        @repeater(.05, 5)
        def x5():
            print "Executing x5"

        import time
        print "Starting Limited Repeating Function"
        token = x5()
        while not token.isSet():
            time.sleep(1)
        print "Expired: x5"

    @param interval      The interval (in seconds) between function invocations.
    @param limit         The limit to the number of function invocations; -1 represents infinity.
    @return              A new decorator with a closure around the original function and the Python threading.
                         Thread used to invoke it.
    """

    def actual_decorator(fn):

        def wrapper(*args, **kwargs):

            class RepeaterTimerThread(threading.Thread):
                def __init__(self):
                    threading.Thread.__init__(self)
                    self._event = threading.Event()

                def run(self):
                    i = 0
                    while i != limit and not self._event.is_set():
                        self._event.wait(interval)
                        fn(*args, **kwargs)
                        i += 1
                    else:
                        if self._event:
                            self._event.set()

                def stopped(self):
                    return not self._event or self._event.is_set()

                def pause(self):
                    self._event.set()

                def resume(self):
                    self._event.clear()

                def stop(self):
                    self._event.set()
                    self.join()

            token = RepeaterTimerThread()
            token.daemon = True
            token.start()
            return token

        return wrapper

    return actual_decorator


def cached(fn):
    cache = {}

    @wraps(fn)
    def wrapper(*args):
        try:
            return cache[args]
        except KeyError:
            rv = fn(*args)
            cache[args] = rv
            return rv

    return wrapper


def add_method(cls):
    """
    This decorator adds a function to given class instance
    https://medium.com/@mgarod/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6
    :param cls: class instance
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)
        setattr(cls, func.__name__, wrapper)
        # Note we are not binding func, but wrapper which accepts self but does exactly the same as func
        return func     # returning func means func can still be used normally
    return decorator


class Singleton(object):
    all_instances = list()

    @staticmethod
    def destroy_all():
        for instance in Singleton.all_instances:
            instance.destroy()

    def __init__(self, cls):
        self.cls = cls
        self.instance = None
        self.all_instances.append(self)

    def destroy(self):
        del self.instance
        self.instance = None

    def __call__(self, *args, **kwargs):
        if self.instance is None:
            self.instance = self.cls(*args, **kwargs)

        return self.instance


class HybridMethod(object):
    """
    Merges a normal method with a classmethod
    First two arguments are (cls, self), where self will match cls if it is a classmethod
    https://stackoverflow.com/a/18078819/2403000
    """

    def __init__(self, fn):
        self._fn = fn

    def __get__(self, obj, cls):
        context = obj if obj is not None else cls

        @wraps(self._fn)
        def hybrid(*args, **kwargs):
            return self._fn(cls, context, *args, **kwargs)

        # Mimic method attributes (not required)
        hybrid.__func__ = hybrid.im_func = self._fn
        hybrid.__self__ = hybrid.im_self = context

        return hybrid



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to cross-platform functionality
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import getpass
import platform
import subprocess


class Platforms(object):
    Windows = 'Windows'
    Linux = 'Linux'
    Mac = 'MacOS'


def get_platform():
    pl = Platforms.Windows

    if 'linux' in sys.platform:
        pl = Platforms.Linux
    elif sys.platform == 'darwin':
        pl = Platforms.Mac

    return pl


def get_home_directory(platform):
    if platform == Platforms.Windows:
        os.environ['TMPDIR'] = os.getenv('TEMP')
        os.environ['HOME'] = os.getenv('HOMEPATH')
        return os.getenv('HOMEPATH')
    if platform == Platforms.Linux:
        return os.getenv('HOME')
    if platform == Platforms.Mac:
        return os.getenv('HOME')


def is_linux():
    """
    Check to see if current platform is Linux
    :return: bool
    """

    platform = get_platform()
    return platform == Platforms.Linux


def is_mac():
    """
    Check to see if current platform is Mac
    :return: bool
    """

    platform = get_platform()
    return platform == Platforms.Mac


def is_windows():
    """
    Check to see if current platform is Windows
    :return: bool
    """

    platform = get_platform()
    return platform == Platforms.Windows


def get_user(lower=True):
    """
    Returns the current user
    :param lower: bool
    :return: str
    """

    username = getpass.getuser()

    return username.lower() if lower else username


def get_permission(filepath):
    """
    Returns the current permission level
    :param filepath: str
    """

    if os.access(filepath, os.R_OK | os.W_OK | os.X_OK):
        return True

    try:
        os.chmod(filepath, 0o775)
        return True
    except Exception:
        return False


def init_env_var(name):
    """
    Initializes a new environment variable if the variable does not exists.
    If it does not exists, nothing happens
    :param name: str, name of the new environment variable
    """

    if name not in os.environ:
        os.environ[name] = ''


def set_env_var(name, value):
    """
    Set the value of an environment variable
    :param name: str, name of the environment variable to set
    :param value: variant, value to initialize environment variable with, empty string by default
    """

    if name not in os.environ:
        init_env_var(name)

    try:
        os.environ[name] = str(value)
    except Exception as e:
        import traceback
        print('{} | {} | name: {} | value: {}'.format(str(e), traceback.format_exc(), name, value))


def get_env_var(name):
    """
    Returns the value of an environment variable
    :param name: str, name of the environment variable
    """

    if name in os.environ:
        return os.environ[name]


def append_env_var(name, value):
    """
    Append string value to the end of the environment variable
    :param name: str, name of the environment variable to set
    :param value: variant, value to initialize environment variable with, empty string by default
    """

    env_value = get_env_var(name=name)

    try:
        env_value += str(value)
    except Exception:
        pass

    set_env_var(name=name, value=env_value)


def get_system_config_directory():
    """
    Returns platform specific configuration directory
    """

    if sys.platform.startswith('darwin'):
        config_directory = os.path.join(os.path.expanduser('~'), 'Library', 'Preferences')
    elif os.name == 'nt':
        config_directory = os.getenv('APPDATA') or os.path.expanduser('~')
    else:
        config_directory = os.getenv('XDG_CONFIG_HOME') or os.path.join(os.path.expanduser('~'), '.config')

    return config_directory


def open_folder(path):
    """
    Open folder using OS default settings
    :param path: str, folder path we want to open
    """

    if sys.platform.startswith('darwin'):
        subprocess.Popen(["open", path])
    elif os.name == 'nt':
        os.startfile(path)
    elif os.name == 'posix':
        subprocess.Popen(["xdg-open", path])
    else:
        raise NotImplementedError('OS not supported: {}'.format(os.name))


def open_file(file_path):
    """
    Open file using OS default settings
    :param file_path: str, file path we want to open
    """

    if sys.platform.startswith('darwin'):
        subprocess.call(('open', file_path))
    elif os.name == 'nt':
        os.startfile(file_path)
    elif os.name == 'posix':
        subprocess.call(('xdg-open', file_path))
    else:
        raise NotImplementedError('OS not supported: {}'.format(os.name))


def machine_info():
    """
    Returns dictionary with information about the current machine
    :return: dict
    """

    machine_dict = {
        'pythonVersion': sys.version,
        'node': platform.node(),
        'OSRelease': platform.release(),
        'OSVersion': platform.platform(),
        'processor': platform.processor(),
        'machineType': platform.machine(),
        'env': os.environ,
        'syspaths': sys.path,
        'executable': sys.executable,
    }

    return machine_dict


def get_architecture():
    """
    Returns architecture of current OS
    :return: str
    """

    if get_platform() == Platforms.Mac:
        if sys.maxsize > 2**32:
            return '64bit'
        else:
            return '32bit'
    else:
        return platform.architecture()[0]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Modules that contains utility functions related with code analysis
"""

from __future__ import print_function, division, absolute_import

import sys
import ast
import string

from tpDcc.libs.python import fileio, path as path_utils


def get_ast_function_args(function_node):
    """
    Returns function arguments from the give AST function node
    :param function_node: ast function node
    :return: list(str)
    """

    found_args = list()

    if not function_node.args:
        return found_args

    defaults = function_node.args.defaults
    args = function_node.args.args
    args.reverse()
    defaults.reverse()
    i = 0
    for arg in args:
        name = arg.id
        if name == 'self':
            continue
        default_value = None
        if i < len(defaults):
            default_value = defaults[i]
        if default_value:
            value = None
            if isinstance(default_value, ast.Str):
                value = "'%s'" % default_value.s
            if isinstance(default_value, ast.Name):
                value = default_value.id
            if isinstance(default_value, ast.Num):
                value = default_value.n
            if value:
                found_args.append('{}={}'.format(name, value))
            else:
                found_args.append(name)
        else:
            found_args.append(name)
        i += 1
    found_args.reverse()

    return found_args


def get_ast_function_name_and_args(function_node):
    """
    Return function name from given AST node
    :param function_node: ast node
    :return: str
    """

    function_name = function_node.name
    found_args = get_ast_function_args(function_node)
    if found_args:
        found_args_name = string.join(found_args, ',')
    else:
        found_args_name = ''

    function_name += '({})'.format(found_args_name)

    return function_name


def get_ast_class_members(class_node, parents=list(), skip_list=None):
    """
    Returns given class node members
    :param class_node: ast class node
    :param parents: list
    :param skip_list: bool
    :return: list(str)
    """

    if skip_list is None:
        skip_list = list()

    class_functions = list()

    for node in class_node.body:
        if isinstance(node, ast.FunctionDef):
            name = node.name
            if skip_list:
                if name in skip_list:
                    continue
            skip_list.append(name)
            name_args = get_ast_function_name_and_args(node)
            if name_args.startswith('__'):
                continue
            name_args = name_args.replace('self,', '')
            class_functions.append(name_args)

    found_parent_functions = list()
    for parent in parents:
        parent_functions = get_ast_class_members(parent, skip_list=skip_list)
        found_parent_functions += parent_functions

    found_parent_functions += class_functions

    return found_parent_functions


def get_ast_class_sub_functions(module_path, class_name):
    """
    Returns list of sub functions for the given class
    :param module_path: str
    :param class_name: str
    :return: list(str)
    """

    defined, defined_dict = get_defined_classes(module_path)
    if not defined:
        return

    if class_name in defined:
        class_node = defined_dict[class_name]
        parents = list()
        bases = class_node.bases
        while bases:
            temp_bases = bases
            find_bases = list()
            for base in temp_bases:
                class_name = None
                if hasattr(base, 'id'):
                    class_name = base.id
                if class_name and class_name in defined_dict:
                    parents.append(defined_dict[class_name])
                    sub_bases = parents[-1].bases
                    if sub_bases:
                        find_bases += sub_bases
            bases = find_bases

        functions = get_ast_class_members(class_node, parents)
        functions.sort()

        return functions


def get_ast_assignment(text, line_number, assignment):
    text = str(text)
    if not text:
        return

    ast_tree = None
    try:
        ast_tree = ast.parse(text, 'string', 'exec')
    except Exception:
        if not ast_tree:
            return

    line_assign_dict = dict()
    value = None

    for node in ast.walk(ast_tree):
        if hasattr(node, 'lineno'):
            current_line_number = node.lineno
            if current_line_number <= line_number:
                if isinstance(node, ast.ImportFrom):
                    for name in node.names:
                        full_name = node.module + '.' + name.name
                        value = ['import', full_name]
                        if not name.asname:
                            line_assign_dict[name.name] = value
                        else:
                            line_assign_dict[name.asname] = ['import', full_name]
                if isinstance(node, ast.Assign):
                    targets = list()
                    if __name__ == '__main__':
                        for target in node.targets:
                            if hasattr(target, 'id'):
                                targets.append(target.id)

                        if hasattr(node.value, ''):
                            pass
                        if hasattr(node.value, 'id'):
                            value = node.value.id
                        if hasattr(node.value, 'func'):
                            value = list()
                            if hasattr(node.value.func, 'value'):
                                if hasattr(node.value.func.value, 'id'):
                                    value.append(node.value.func.value.id)
                                    value.append(node.value.func.attr)

                        if targets:
                            for target in targets:
                                if value:
                                    line_assign_dict[target] = value

            if current_line_number > line_number:
                continue

    return line_assign_dict


def get_package_path_from_name(module_name, return_module_path=False):
    """
    Returns package path from given package name
    :param module_name: str, name of the module we want to retrieve path of
    :param return_module_path: bool
    :return: str
    """

    split_name = module_name.split('.')
    if len(split_name) > 1:
        sub_path = string.join(split_name[:-1], '/')
    else:
        sub_path = module_name

    paths = sys.path

    found_path = None
    for path in paths:
        test_path = path_utils.join_path(path, sub_path)
        if path_utils.is_dir(test_path):
            found_path = path

    if not found_path:
        return None

    test_path = found_path
    good_path = ''
    inc = 0

    for name in split_name:
        if inc == len(split_name) - 1:
            if return_module_path:
                good_path = path_utils.join_path(good_path, '{}.py'.format(name))
                break

        test_path = path_utils.join_path(test_path, name)
        if not path_utils.is_dir(test_path):
            continue
        files = fileio.get_files(test_path)
        if '__init__.py' in files:
            good_path = test_path
        else:
            return None

        inc += 1

    return good_path


def get_line_imports(lines):
    """
    Returns all import lines in the given lines
    :param lines: list(str)
    :return: dict(str, str)
    """

    # TODO: Replace this function by AST implementation

    module_dict = dict()

    for line in lines:
        line = str(line)
        split_line = line.split()
        split_line_count = len(split_line)
        for i in range(split_line_count):
            module_prefix = ''
            if split_line[i] == 'import':
                if i > 1:
                    if split_line[i - 2] == 'from':
                        module_prefix = split_line[i - 1]
                if i < split_line_count - 1:
                    module = split_line[i + 1]
                    namespace = module
                    if module_prefix:
                        module = '{}.{}'.format(module_prefix, module)
                    module_path = get_package_path_from_name(module, return_module_path=True)
                    module_dict[namespace] = module_path

    return module_dict


def get_defined(module_path, name_only=False):
    """
    Get classes and definitions from the text of a module
    """

    file_text = fileio.get_file_text(module_path)

    functions = list()
    classes = list()

    ast_tree = ast.parse(file_text, 'string', 'exec')
    for node in ast_tree.body:
        found_args_name = ''
        if isinstance(node, ast.FunctionDef):
            function_name = node.name
            if not name_only:
                function_name = get_ast_function_name_and_args(node)
            functions.append(function_name)
        if isinstance(node, ast.ClassDef):
            class_name = node.name + '()'
            for sub_node in node.body:
                if isinstance(sub_node, ast.FunctionDef):
                    if sub_node.name == '__init__':
                        found_args = get_ast_function_args(sub_node)
                        if found_args:
                            found_args_name = string.join(found_args, ',')
                        else:
                            found_args_name = ''
                        class_name = '{}({})'.format(node.name, found_args_name)
            classes.append(class_name)

    classes.sort()
    functions.sort()
    defined = classes + functions

    return defined


def get_defined_classes(module_path):
    """
    Returns classes of the given module
    :param module_path: str
    :return: list(str)
    """

    file_text = fileio.get_text(module_path)

    defined = list()
    defined_dict = dict()

    if not file_text:
        return None, None

    ast_tree = ast.parse(file_text)
    for node in ast_tree.body:
        if isinstance(node, ast.ClassDef):
            defined.append(node.name)
            defined_dict[node.name] = node

    return defined, defined_dict



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with time and date
"""

from __future__ import print_function, division, absolute_import

import time
from datetime import datetime


def convert_number_to_month(month_int):
    """
    Return a month as string given a month number
    :param month_int: int
    :return: str
    """

    months = ['January',
              'February',
              'March',
              'April',
              'May',
              'June',
              'July',
              'August',
              'September',
              'October',
              'November',
              'December']

    month_int -= 1
    if month_int < 0 or month_int > 11:
        return

    return months[month_int]


def get_current_time(date_and_time=True, reverse_date=False):
    """
    Returns current time
    :param date_and_time: bool, Whether to return only the time or time and data
    :param reverse_date: bool, Whether to return date with {year}-{month}-{day} format or {day}-{month}-{year} format
    :return: str
    """

    mtime = time.time()
    date_value = datetime.fromtimestamp(mtime)
    hour = str(date_value.hour)
    minute = str(date_value.minute)
    second = str(int(date_value.second))

    if len(hour) == 1:
        hour = '0' + hour
    if len(minute) == 1:
        minute = '0' + minute
    if len(second) == 1:
        second += '0'

    time_value = '{}:{}:{}'.format(hour, minute, second)

    if not date_and_time:
        return time_value
    else:
        year = date_value.year
        month = date_value.month
        day = date_value.day

        if reverse_date:
            return '{}-{}-{} {}'.format(year, month, day, time_value)
        else:
            return '{}-{}-{} {}'.format(day, month, year, time_value)


def get_current_date(reverse_date=False):
    """
    Returns current date
    :param reverse_date: bool, Whether to return date with {year}-{month}-{day} format or {day}-{month}-{year} format
    :return: str
    """

    mtime = time.time()
    date_value = datetime.fromtimestamp(mtime)
    year = date_value.year
    month = date_value.month
    day = date_value.day

    if reverse_date:
        return '{}-{}-{}'.format(year, month, day)
    else:
        return '{}-{}-{}'.format(day, month, year)


def time_ago(timestamp):
    """
    Returns a pretty string for how long ago the given timestamp was
    Example:
        print timeAgo("2017-06-06 01:56:00"
        # 2 years ago
    :param timestamp: str
    :return:str
    """

    t1 = int(timestamp)
    t1 = datetime.fromtimestamp(t1)

    t2 = datetime.now()
    diff = t2 - t1

    day_diff = diff.days
    seconds_diff = diff.seconds

    if day_diff < 0:
        return ''

    if day_diff == 0:
        if seconds_diff < 10:
            return 'just now'
        if seconds_diff < 60:
            return str(seconds_diff) + ' seconds ago'
        if seconds_diff < 120:
            return 'a minute ago'
        if seconds_diff < 3600:
            return str(seconds_diff / 60) + " minutes ago"
        if seconds_diff < 7200:
            return 'an hour ago'
        if seconds_diff < 86400:
            return str(seconds_diff / 3600) + ' hours ago'

    if day_diff == 1:
        return 'yesterday'

    if day_diff < 7:
        return str(day_diff) + ' days ago'

    if day_diff < 31:
        v = day_diff / 7
        if v == 1:
            return str(v) + ' week ago'
        return str(day_diff / 7) + ' weeks ago'

    if day_diff < 365:
        v = day_diff / 30
        if v == 1:
            return str(v) + ' month ago'
        return str(v) + ' months ago'

    v = day_diff / 365
    if v == 1:
        return str(v) + ' year ago'
    return str(v) + ' years ago'



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with debugging purposes
"""


from __future__ import print_function, division, absolute_import


def format_message(fname, expected, actual, flag):
    """
    Convenience function that returns nicely formatted error/warning messages
    """

    def _format(types):
        return ', '.join([str(t).split("'")[1] for t in types])

    expected, actual = _format(expected), _format(actual)
    msg = "'{}' method ".format(fname) + ('accepts', 'returns')[flag] + ' ({}), but '.format(expected) + \
          ('was given', 'result is')[flag] + ' ({})'.format(actual)
    return msg


def debug_object_string(obj, msg):
    """
    Returns a debug string depending of the type of the object
    :param obj: Python object
    :param msg: message to log
    :return: str, debug string
    """

    import inspect
    # debug a module
    if inspect.ismodule(obj):
        return '[%s module] :: %s' % (obj.__name__, msg)

    # debug a class
    elif inspect.isclass(obj):
        return '[%s.%s class] :: %s' % (obj.__module__, obj.__name__, msg)

    # debug an instance method
    elif inspect.ismethod(obj):
        return '[%s.%s.%s method] :: %s' % (obj.im_class.__module__, obj.im_class.__name__, obj.__name__, msg)

    # debug a function
    elif inspect.isfunction(obj):
        return '[%s.%s function] :: %s' % (obj.__module__, obj.__name__, msg)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions to handle Python modules
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import imp
import inspect
import importlib

from tpDcc.libs import python
from tpDcc.libs.python import path as path_utils


def is_dotted_module_path(module_path):
    """
    Returns whether given module path is a dotted one (tpDcc.libs.python.modules) or not
    :param module_path: str
    :return: bool
    """

    return len(module_path.split('.')) > 2


def convert_to_dotted_path(path):
    """
    Returns a dotted path relative to the given path
    :param path: str, (eg. randomPath/folder/test.py)
    :return: str, dotted path (eg. folder.test)
    """

    directory, file_path = os.path.split(path)
    directory = path_utils.clean_path(directory)
    file_name = os.path.splitext(file_path)[0]
    package_path = [file_name]
    sys_path = [path_utils.clean_path(p) for p in sys.path]
    drive_letter = os.path.splitdrive(path)[0] + '\\'
    while directory not in sys_path:
        directory, name = os.path.split(directory)
        directory = path_utils.clean_path(directory)
        if directory == drive_letter or name == '':
            return ''
        package_path.append(name)

    return '.'.join(reversed(package_path))


def import_module(module_path, name=None):
    """
    Imports the given module path. If the given module path is a dotted one, import lib will be used. Otherwise, it's
    expected that given module path is the absolute path to the source file. If name argument is not given, then the
    basename without the extension will be used
    :param module_path: str, module path. Can be a dotted path (tpDcc.libs.python.modules) or an absolute one
    :param name: str, name for the imported module which will be used if the module path is an absolute path
    :return: ModuleObject, imported module object
    """

    if is_dotted_module_path(module_path) and not os.path.exists(module_path):
        try:
            return importlib.import_module(module_path)
        except ImportError:
            python.logger.error('Failed to load module: "{}"'.format(module_path), exc_info=True)

    try:
        if os.path.exists(module_path):
            if not name:
                name = os.path.splitext(os.path.basename(module_path))[0]
            if name in sys.modules:
                return sys.modules[name]
        if os.path.isdir(module_path):
            module_path = os.path.join(module_path, '__init__.py')
            if not os.path.exists(module_path):
                raise ValueError('Cannot find module path: "{}"'.format(module_path))
        return imp.load_source(name, os.path.realpath(module_path))
    except ImportError:
        python.logger.error('Failed to load module: "{}"'.format(module_path))
        raise


def iterate_modules(path, exclude=None):
    """
    Iterates all the modules of the given path
    :param path: str, folder path to iterate
    :param exclude: list(str), list of files to exclude
    :return: iterator
    """

    if not exclude:
        exclude = list()

    _exclude = ['__init__.py', '__init__.pyc']
    for root, dirs, files in os.walk(path):
        if '__init__.py' not in files:
            continue
        for f in files:
            base_name = os.path.splitext(f)[0]
            if f not in _exclude and base_name not in exclude:
                module_path = os.path.join(root, f)
                if f.endswith('.py') or f.endswith('.pyc'):
                    yield module_path


def iterate_module_members(module_to_iterate, predicate=None):
    """
    Iterates all the members of the given modules
    :param module_to_iterate: ModuleObject, module object to iterate members of
    :param predicate: inspect.cass, if given members will be restricted to given inspect class
    :return: iterator
    """

    for mod in inspect.getmembers(module_to_iterate, predicate=predicate):
        yield mod


def iterate_module_subclasses(module, class_type):
    """
    Iterates all classes within a module object returning all subclasses of given type
    :param module: ModuleObject, module object to iterate subclasses of
    :param class_type: object, class object to find
    :return: generator(object), generator function returning class objects
    """

    for member in iterate_module_members(module, predicate=inspect.isclass):
        if issubclass(member, class_type):
            yield member



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to handle version files
"""

import getpass

from tpDcc.libs.python import folder, path, fileio, sort, name as name_utils


class SemanticVersion(object):
    """
    Class that defines Semantic version following: <https://semver.org>
    """

    def __init__(self, major, minor, patch):
        super(SemanticVersion, self).__init__()

        assert(isinstance(major, int))
        assert(isinstance(minor, int))
        assert(isinstance(patch, int))
        self._major = major
        self._minor = minor
        self._patch = patch

    @staticmethod
    def from_string(version_string):
        """
        Returns semantic version instance from given string
        :param version_string: str
        :return: SemanticVersion
        """

        major, minor, patch = version_string.split('.')
        return SemanticVersion(int(major), int(minor), int(patch))

    def __str__(self):
        return '{}.{}.{}'.format(self._major, self._minor, self._patch)

    def __eq__(self, other):
        return all([self.major == other.major, self.minor == other.minor, self.patch == other.patch])

    def __ge__(self, other):
        lhs = int(''.join([str(self.major), str(self.minor), str(self.patch)]))
        rhs = int(''.join([str(other.major), str(other.minor), str(other.patch)]))
        return lhs >= rhs

    def __gt__(self, other):
        lhs = int(''.join([str(self.major), str(self.minor), str(self.patch)]))
        rhs = int(''.join([str(other.major), str(other.minor), str(other.patch)]))
        return lhs > rhs

    @property
    def major(self):
        """
        Returns major field of the semantic version
        :return: int
        """

        return self._major

    @property
    def minor(self):
        """
        Returns minor field of the semantic version
        :return: int
        """

        return self._minor

    @property
    def patch(self):
        """
        Returns patch field of the semantic version
        :return: int
        """

        return self._patch


class VersionFile(object):
    """
    Utility class to hold version for files and folders
    """

    def __init__(self, file_path):
        self._file_path = file_path
        self._path = path.get_dirname(file_path)
        self._version_folder_name = '__version__'
        self._version_name = 'version'
        self._version_folder = None
        self._comment_file = None
        self._updated_old = False

    @property
    def file_path(self):
        return self._file_path

    @property
    def file_name(self):
        if not self._file_path:
            return ''

        return path.get_basename(self._file_path)

    @property
    def updated_old(self):
        return self._updated_old

    def get_version_path(self, version_number):
        """
        Returns the path to the version
        :param version_number: int, version number
        :return: str, path to the version
        """

        return self._get_version_path(version_number)

    def get_version_comment(self, version_number):
        """
        Returns the comment of the given version
        :param version_number: int, version number
        :return: str, version comment
        """

        comment, user = self.get_version_data(version_number)

        return comment

    def set_version_folder(self, folder_path):
        """
        Sets the folder where the version folder should be created
        :param folder_path: str, full path to where the version folder should be created
        """

        self._path = folder_path

    def set_version_folder_name(self, folder_name):
        """
        Sets the name of the version folder
        :param folder_name: str, name of the version folder
        """

        self._version_folder_name = folder_name

    def set_version_name(self, version_name):
        """
        Sets the version name
        :param version_name: str, name of the version
        """

        self._version_name = version_name

    def has_default(self):
        if not self._version_name:
            return

        file_name = self._default_version_file_name()
        if path.is_file(file_name):
            return True

        return False

    def has_versions(self):
        """
        Returns if the current version file has version or not
        :return: bool
        """

        version_folder = self._get_version_folder()
        if path.is_dir(version_folder):
            return True

        return False

    def get_default(self):
        file_name = self._default_version_file_name()

        return file_name

    def get_count(self):
        """
        Returns the number of versions
        :return: int
        """

        versions = self.get_version_numbers()
        if versions:
            return len(versions)

        return 0

    def save_comment(self, comment=None, version_file=None):
        """
        Saves a comment to a log file
        :param comment: str, comment to save
        :param version_file: str, correspnding version file
        """

        version = version_file.split('.')
        if version:
            version = version[-1]

        user = getpass.getuser()

        if not comment:
            comment = '-'
        comment.replace('"', '\"')

        comment_file = fileio.FileWriter(file_path=self._comment_file)
        comment_file.set_append(True)
        comment_file.write(['version = {}; comment = "{}"; user = "{}"'.format(version, comment, user)])
        comment_file.close_file()

    def save(self, comment=None):
        """
        Saves a new version
        :param comment: str, comment to add to the version
        :return: str, new version file name
        """

        self._prepare_directories()

        if comment is None:
            comment = ' '
        else:
            comment.replace('\n', '   ')
            comment.replace('\r', '   ')

        unique_file_name = self._increment_version_file_name()
        self._save(unique_file_name)
        self.save_comment(comment, unique_file_name)

        return unique_file_name

    def save_default(self):
        self._prepare_directories()

        file_name = self._default_version_file_name()
        if file_name:
            self._save(file_name)
        else:
            print('Could not save default version!')

        return file_name

    def get_latest_version(self):
        """
        Returns the file path to the latest version
        :return: str
        """

        versions = self.get_versions()
        latest_version = versions[-1]

        return path.join_path(self._file_path, '{}/{}'.format(self._version_folder_name, latest_version))

    def get_versions(self, return_version_numbers=False):
        """
        Get file paths of all versions
        :param return_version_numbers: bool, Whether to return also version numbers or only paths
        :return: variant, list<str> | list<str>, list<int>
        """

        version_folder = self._get_version_folder()
        version_files = folder.get_files_and_folders(version_folder)
        if not version_files:
            print('WARNING: Impossible to get versions because no version exist!')
            return None

        number_list = list()
        pass_files = list()
        for f in version_files:
            if not f.startswith(self._version_name):
                continue
            split_name = f.split('.')
            if not len(split_name) == 2:
                continue

            version_number = int(split_name[1])
            number_list.append(version_number)
            pass_files.append(f)

        if not pass_files:
            print('WARNING: No valid version files found in folder: {}'.format(version_folder))
            return

        quick_sort = sort.QuickNumbersListSort(number_list)
        quick_sort.set_follower_list(pass_files)
        pass_files = quick_sort.run()

        pass_dict = dict()
        for i in range(len(number_list)):
            pass_dict[pass_files[0][i]] = pass_files[1][i]

        if return_version_numbers:
            return pass_dict, pass_files[0]
        else:
            return pass_dict

    def get_version_numbers(self):
        """
        Returns numbers of all versions
        :return: list<int>
        """

        version_folder = self._get_version_folder()
        version_files = folder.get_files_and_folders(version_folder)
        if not version_files:
            print('WARNING: Impossible to get version numbers because no version exist!')
            return None

        number_list = list()
        for f in version_files:
            if not f.startswith(self._version_name):
                continue
            split_name = f.split('.')
            if split_name[1] == 'json' or split_name[1] == 'default' or not len(split_name) == 2:
                continue

            version_number = int(split_name[1])
            number_list.append(version_number)

        number_list.sort()

        return number_list

    def get_version_data(self, version_number):
        """
        Returns the data (comment and user) of the given version
        :param version_number: int
        :return: tuple(str, str)
        """

        comment_path = self._get_comment_path()
        if not comment_path:
            return None, None

        if path.is_file(comment_path):
            read = fileio.FileReader(comment_path)
            lines = read.read()
            version = None
            comment = None
            user = None
            for line in lines:
                start_index = line.find('"')
                if start_index > -1:
                    end_index = line.find('";')
                    sub_part = line[start_index + 1:end_index]
                    sub_part = sub_part.replace('"', '\\"')
                    line = line[:start_index + 1] + sub_part + line[end_index:]

                # Get version, comment and user variables by executing the line
                try:
                    exec(line)
                except Exception:
                    pass

                if version == version_number:
                    return comment, user

        return None, None

    def get_organized_version_data(self):
        """
        Returns all the version data:
        [version, comment, user, file_size, modified, version_file]
        :return: list
        """

        versions = self.get_versions(return_version_numbers=True)
        if not versions:
            return

        version_paths = versions[0]
        version_numbers = versions[1]

        comment_path = self._get_comment_path()
        if not comment_path:
            return []

        datas = list()
        if path.is_file(comment_path):
            read = fileio.FileReader(comment_path)
            lines = read.read()
            for l in lines:
                line_info_dict = dict()
                user = None
                comment = None
                split_line = l.split(';')
                for sub_line in split_line:
                    assign = sub_line.split('=')
                    if assign and assign[0]:
                        data_name = assign[0].strip()
                        data_value = assign[1].strip()
                        line_info_dict[data_name] = data_value

                if 'version' not in line_info_dict:
                    continue

                version = int(line_info_dict['version'])
                if not int(line_info_dict['version']) in version_numbers:
                    continue

                if 'comment' in line_info_dict:
                    comment = line_info_dict['comment']
                    comment = comment[1:-1]
                if 'user' in line_info_dict:
                    user = line_info_dict['user']
                    user = user[1:-1]

                version_file = version_paths[version]
                version_file = path.join_path(self._file_path, '{}/{}'.format(self._version_folder_name, version_file))

                file_size = fileio.get_file_size(version_file)
                modified = fileio.get_last_modified_date(version_file)

                datas.append([version, comment, user, file_size, modified, version_file])

        return datas

    def _prepare_directories(self):
        """
        Internal function used to prepare necessary directories and files for version
        :return:
        """
        self._create_version_folder()
        self._create_comment_file()

    def _create_version_folder(self):
        self._version_folder = folder.create_folder(self._version_folder_name, self._path)

    def _create_comment_file(self):
        self._comment_file = fileio.create_file('comments.txt', self._version_folder)

    def _increment_version_file_name(self):
        version_path = path.join_path(self._version_folder, self._version_name + '.1')
        return path.unique_path_name(version_path)

    def _get_version_number(self, file_path):
        version_number = name_utils.get_end_number(file_path)
        return version_number

    def _get_version_path(self, version_number):
        version_path = path.join_path(self._get_version_folder(), self._version_name + '.' + str(version_number))
        return version_path

    def _get_version_folder(self):
        if path.is_file(self._file_path):
            version_dir = path.get_dirname(self._file_path)
            version_path = path.join_path(version_dir, self._version_folder_name)
        else:
            version_path = path.join_path(self._file_path, self._version_folder_name)

        return version_path

    def _get_comment_path(self):
        version_folder = self._get_version_folder()
        comment_path = None
        if version_folder:
            comment_path = path.join_path(version_folder, 'comments.txt')

        return comment_path

    def _default_version_file_name(self):
        if not self._version_name:
            return

        version_folder = self._get_version_folder()
        version_path = path.join_path(version_folder, self._version_name + '.default')

        return version_path

    def _save(self, file_name):
        self._prepare_directories()
        if path.is_dir(self._file_path):
            folder.copy_folder(self._file_path, file_name)
        elif path.is_file(self._file_path):
            fileio.copy_file(self._file_path, file_name)

    def delete_version(self, version_number):
        """
        Deletes specific version file
        :param version_number: int
        """

        version_path = self.get_version_path(version_number)
        if path.is_file(version_path):
            fileio.delete_file(version_path)
        else:
            folder.delete_folder(path)


def delete_version(folder, keep=1):
    """
    Deletes all version in the given folder maintaining only the exact number of keep versions
    :param folder: str
    :param keep: int
    """

    version_inst = VersionFile(folder)
    version_list = version_inst.get_version_numbers()
    if not version_list:
        return

    count = len(version_list)
    if count <= keep:
        return

    deleted = 0
    for version in version_list:
        version_inst.delete_version(version)
        deleted += 1
        if count - deleted == keep:
            break



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to profile Python code
"""

from __future__ import print_function, division, absolute_import

import time
import pstats
import cProfile as profile
from functools import wraps
from collections import defaultdict


def profile_function(sort_key='time', rows=30):
    def _(fn):
        @wraps(_)
        def __(*fargs, **fkwargs):
            prof = profile.Profile()
            ret = prof.runcall(f, *fargs, **fkwargs)
            pstats.Stats(prof).strip_dirs().sort_stats(sort_key).print_stats(rows)
            return ret
        return __
    return _


class LapCounter():
    LapTimes = 0
    LapList = list()

    def __init__(self):
        current_time = time.time()
        self._all_start = current_time
        self._start = current_time
        self._end = current_time

    def count(self, string=''):
        self._end = time.time()
        lap_str = 'lap_time : ', string, self.LapTimes, ':', self._end - self._start
        self.LapList.append(lap_str)
        self.LapTimes += 1
        self._start = time.time()

    def lap_print(self, print_flag=True, window=None):
        total_time = time.time() - self._all_start
        if window:
            out_time = '{:.5f}'.format(total_time)
            try:
                window.time_label.setText('- Calculation Time - ' + out_time + ' sec')
            except Exception as e:
                pass

        if print_flag:
            print('----------------------------------')
            for lap_time in self.LapList:
                print(lap_time)
            print('Total time : {}'.format(total_time))

    def reset(self):
        self._all_start = time.time()
        self._start = time.time()
        self.LapList = list()


class IntegrationCounter():
    def __init__(self):
        current_time = time.time()
        self._all_start = current_time
        self._start = current_time
        self._end = current_time
        self._integration_dict = defaultdict(lambda: 0)

    def count(self, string=''):
        self._end = time.time()
        self._integration_dict[string] += self._end - self._start
        self._start = time.time()

    def integration_print(self):
        for string, integration in self._integration_dict.items():
            print('Integration time : ', string, integration)

    def reset(self):
        self._all_start = time.time()
        self._start = time.time()
        self._integration_dict = defaultdict(lambda: 0)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different types of timers
"""

from __future__ import print_function, division, absolute_import

import sys
import time


class StopWatch(object):
    """
    Class that can be used to check how long a command takes to run
    """

    running = 0

    def __init__(self):
        self.time = None
        self.feedback = True

    def __del__(self):
        self.end()

    def start(self, description='', feedback=True):
        self.feedback = feedback
        if feedback:
            tabs = '\t' * self.running
            sys.utils_log.debug('{}started timer: {}'.format(tabs, description))
        self.time = time.time()
        if feedback:
            self.__class__.running += 1

    def end(self):
        if not self.time:
            return

        seconds = time.time() - self.time
        self.time = None

        seconds = round(seconds, 2)
        minutes = None
        if seconds > 60:
            minutes, seconds = divmod(seconds, 60)
            seconds = round(seconds, 2)
            minutes = int(minutes)

        if self.feedback:
            tabs = '\t' * self.running
            if minutes is None:
                sys.utils_log.debug('{}end timer: {} seconds'.format(tabs, seconds))
            else:
                if minutes > 1:
                    sys.utils_log.debug('{} end timer: {}  minutes, {} seconds'.format(tabs, minutes, seconds))
                elif minutes == 1:
                    sys.utils_log.debug('{} end timer: {} minute, {} seconds'.format(tabs, minutes, seconds))
            self.__class__.running -= 1

        return minutes, seconds

    def stop(self):
        return self.end()



# !/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

"""
Module that contains classes to create basic logs
"""

import os
import logging
from logging.handlers import RotatingFileHandler, SysLogHandler

try:
    import curses
except ImportError:
    curses = None

loggers = []

from tpDcc.libs.python import osplatform


class LoggerLevel:
    def __init__(self):
        pass

    INFO = logging.INFO
    WARNING = logging.WARNING
    DEBUG = logging.DEBUG
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class Logger(object):

    _LOGGER_IDENTIFIER = '_is_tpdcc_logger'
    _LOGGER_DEFAULT_NAME = 'tpdcc_logger'
    _LOGGER_INTERNAL_HANDLER_IS_CUSTOM_LOG_LEVEL = '_is_tpdcc_logger_internal_handler_custom_log_level'
    _LOGGER_DEFAULT_FORMAT = '[%(levelname)1.1s %(asctime)s | %(name)s | %(module)s:%(lineno)d]: %(message)s'
    _LOGGER_DEFAULT_DATE_FORMAT = '%d/%m/%y %H:%M:%S'

    def __init__(self, name=None, log_file=None, level=logging.DEBUG, formatter=None, max_bytes=0, backup_count=0,
                 file_log_level=None, enable_stderr_logger=True):
        super(Logger, self).__init__()

        self._name = name or self._LOGGER_DEFAULT_NAME
        self._log_level = level or logging.DEBUG
        self._log_file = log_file
        self._formatter = formatter

        self._logger = Logger.create_logger(name=self._name, log_file=self._log_file, level=self._log_level,
                                            formatter=self._formatter, max_bytes=max_bytes,
                                            file_log_level=file_log_level, backup_count=backup_count,
                                            enable_stderr_logger=enable_stderr_logger)

    def get_name(self):
        return self._name

    def get_logger(self):
        return self._logger

    name = property(get_name)
    logger = property(get_logger)

    @staticmethod
    def create_logger(name, log_file=None, level=logging.DEBUG, formatter=None, max_bytes=0, backup_count=0,
                      file_log_level=None, enable_stderr_logger=True):
        """
        Creates a fully configured logger instance
        :param name: str, name of the logger object
        :param log_file: str, If given, path where the log should be stored
        :param level: int, minimum logging level to display
        :param formatter: Python logging Formater object to apply
        :param max_bytes: int, size of log file when rollover should occur. Default is 0, so rollover never should occur
        :param backup_count: int, number of backups to keep. Default is 0, so no backups are done
        :param file_log_level: int, logging level for the file logger
        :param enable_stderr_logger: bool, Whether the default stderr logger be disabled or not
        :return: Logger object
        """

        if formatter is None:
            formatter = logging.Formatter(fmt=Logger._LOGGER_DEFAULT_FORMAT, datefmt=Logger._LOGGER_DEFAULT_DATE_FORMAT)

        _logger = logging.getLogger(name)
        _logger.propagate = False
        _logger.setLevel(level)

        stderr_stream_handler = None
        for handler in list(_logger.handlers):
            if hasattr(handler, Logger._LOGGER_IDENTIFIER):
                if isinstance(handler, logging.FileHandler):
                    _logger.removeHandler(handler)
                    continue
                elif isinstance(handler, logging.StreamHandler):
                    stderr_stream_handler = handler

            handler.setLevel(level)
            handler.setFormatter(formatter)

        if not enable_stderr_logger:
            if stderr_stream_handler is not None:
                _logger.removeHandler(stderr_stream_handler)
        elif stderr_stream_handler is None:
            stderr_stream_handler = logging.StreamHandler()
            setattr(stderr_stream_handler, Logger._LOGGER_IDENTIFIER, True)
            stderr_stream_handler.setLevel(level)
            stderr_stream_handler.setFormatter(formatter)
            _logger.addHandler(stderr_stream_handler)

        if log_file:
            rotating_file_handler = RotatingFileHandler(filename=log_file, maxBytes=max_bytes, backupCount=backup_count)
            setattr(rotating_file_handler, Logger._LOGGER_IDENTIFIER, True)
            rotating_file_handler.setLevel(file_log_level or level)
            rotating_file_handler.setFormatter(formatter)
            _logger.addHandler(rotating_file_handler)

        return _logger

    def reset(self):
        """
        Resets the internal logger to its initial configuration
        """

        self._logger = Logger.create_logger(name=self._name, log_file=self._log_file, level=self._log_level,
                                            formatter=self._formatter)

    def log_level(self, level=logging.DEBUG, update_custom_handlers=False):
        """
        Set the minimum log level for the internal logger
        :param level: int, new miminimum log level
        :param update_custom_handlers: bool, Whether to only reconfigure internal logger handlers only or also
            reconfigure custom handlers
        """

        self._logger.setLevel(level)

        # Reconfigure existing internal handlers
        for handler in list(self._logger.handlers):
            if hasattr(handler, Logger._LOGGER_IDENTIFIER) or update_custom_handlers:
                # If handler is a custom one we do not update log level
                if hasattr(handler, Logger._LOGGER_INTERNAL_HANDLER_IS_CUSTOM_LOG_LEVEL):
                    continue

                # Updates the log level for all default handlers
                handler.setLevel(level)

        self._log_level = level

    def formatter(self, formatter, update_custom_handlers=False):
        """
        Set the formatter for all handlers of the default logger
        :param formatter: formatter, new Python logging formatter
        :param update_custom_handlers: bool, Whether to only reconfigure internal logger handlers only or also
            reconfigure custom handlers
        """

        for handler in list(self._logger.handlers):
            if hasattr(handler, Logger._LOGGER_IDENTIFIER) or update_custom_handlers:
                handler.setFormatter(formatter)

        self._formatter = formatter

    def log_file(self, file_name, formatter=None, mode='a', max_bytes=0, backup_count=0, encoding=None,
                 log_level=None, enable_stderr_logger=True):
        """
        Setup logging to a file (using a RotatingFileHandler)
        :param file_name: str, file name of the log file
        :param formatter: Python logging formatter
        :param mode: str, mode to open the file with
        :param max_bytes: int, size of the log file when rollover should occurs. If 0, rollover never occurs
        :param backup_count: int, number of backups to keep. If 0, backup never occurs
        :param encoding: str, used to open the file with that encoding
        :param log_level: int, set a custom log level for the file logger, else uses the current global log level
        :param enable_stderr_logger: bool, Whether the default stderr logger be enabled or not
        """

        # If an internal RotatingFileHandler already exists we remove it
        self._remove_internal_loggers(enable_stderr_logger=enable_stderr_logger)

        if formatter is None and self._formatter is None:
            formatter = logging.Formatter(fmt=Logger._LOGGER_DEFAULT_FORMAT, datefmt=Logger._LOGGER_DEFAULT_DATE_FORMAT)

        # If want, we add a new RotatingFileHandler
        if file_name:
            rotating_file_handler = RotatingFileHandler(file_name, mode=mode, maxBytes=max_bytes,
                                                        backupCount=backup_count, encoding=encoding)
            setattr(rotating_file_handler, Logger._LOGGER_IDENTIFIER, True)
            if log_level:
                setattr(rotating_file_handler, Logger._LOGGER_INTERNAL_HANDLER_IS_CUSTOM_LOG_LEVEL, True)

            # Configure the handler and add it to the logger
            rotating_file_handler.setLevel(log_level or self._log_level)
            rotating_file_handler.setFormatter(formatter or self._formatter)
            self._logger.addHandler(rotating_file_handler)

    def sys_log(self, facility=SysLogHandler.LOG_USER, enable_stderr_logger=False):
        """
        Setup logging to sys log and disable other internal logger handlers
        :param facility: syslog facility to log to
        :param enable_stderr_logger: bool, Whether the default stderr logger be enabled or not
        :return: SysLogHandler
        """

        self._remove_internal_loggers(enable_stderr_logger=enable_stderr_logger)

        sys_log_handler = SysLogHandler(facility=facility)
        setattr(sys_log_handler, Logger._LOGGER_IDENTIFIER, True)
        self._logger.addHandler(sys_log_handler)

        return sys_log_handler

    def start_temp_log(self):
        """
        Initializes a new temp and stores its results in environment variable
        """

        start_temp_log(self.name)

    def record_temp_log(self, value):
        """
        Adds a new value to the temp log with the given name (if exists)
        :param value: str
        """

        record_temp_log(self.name, value)

    def end_temp_log(self):
        """
        Removes temp log with given name and returns its contents
        :return: str
        """

        return end_temp_log(self.name)

    def _remove_internal_loggers(self, enable_stderr_logger=False):
        """
        Removes the internal logger handlers of the internal logger
        :param enable_stderr_logger: bool, Whether the default stderrl logger should be disabled or not
        """

        for handler in list(self._logger.handlers):
            if hasattr(handler, Logger._LOGGER_IDENTIFIER):
                if isinstance(handler, RotatingFileHandler):
                    self._logger.removeHandler(handler)
                elif isinstance(handler, SysLogHandler):
                    self._logger.removeHandler(handler)
                elif isinstance(handler, logging.StreamHandler) and not enable_stderr_logger:
                    self._logger.removeHandler(handler)


def create_logger(logger_name, logger_path):
    """
    Function to create standard logger for modules and apps
    :param logger_name: str, name of the logger
    :return: Logger
    """

    # If the given path does not exists we create the given folder
    if not os.path.exists(logger_path):
        os.makedirs(logger_path)

    log_file = os.path.join(logger_path, logger_name + '.log')
    logger = Logger(name=logger_name, log_file=log_file)
    global loggers
    loggers.append(logger)

    return logger


def get_logger(logger_name):
    """
    Returns a logger, if exists, from the list of already created loggers
    :param logger_name: str, name of the logger we want to retrieve
    :return: variant, Logger || None
    """

    global loggers
    for l in loggers:
        if l.name == logger_name:
            return l

    return None


def start_temp_log(log_name):
    """
    Initializes a new temp and stores its results in environment variable
    :param log_name: str, name of the log
    """

    osplatform.set_env_var('{}_KEEP_TEMP_LOG'.format(log_name.upper()), 'True')
    osplatform.set_env_var('{}_TEMP_LOG'.format(log_name.upper()), '')


def record_temp_log(log_name, value):
    """
    Adds a new value to the temp log with the given name (if exists)
    :param log_name: str, name of the log we want to add value into
    :param value: str
    """

    if osplatform.get_env_var('{}_KEEP_TEMP_LOG'.format(log_name.upper())) == 'True':
        value = value.replace('\t', '  ') + '\n'
        osplatform.append_env_var('{}_TEMP_LOG'.format(log_name.upper()), value)


def end_temp_log(log_name):
    """
    Removes temp log with given name and returns its contents
    :param log_name: str, nam of the temp log we want to remove
    :return: str
    """

    osplatform.set_env_var('{}_KEEP_TEMP_LOG'.format(log_name.upper()), 'False')
    value = osplatform.get_env_var('{}_TEMP_LOG'.format(log_name.upper()))
    osplatform.set_env_var('{}_TEMP_LOG'.format(log_name.upper()), '')
    osplatform.set_env_var('{}_LAST_TEMP_LOG'.format(log_name.upper()), value)

    return value
